/******************************************************************************
 *
 * Mass-conserving ProblemStat for AMDiS
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis/trunk/extensions
 *
 * Author: Simon Praetorius
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

/** \file ProblemStatMassConserve_mm.h */


#ifndef AMDIS_EXTENSIONS_PROBLEM_STAT_MASS_CONSERVE_MULTI_MESH
#define AMDIS_EXTENSIONS_PROBLEM_STAT_MASS_CONSERVE_MULTI_MESH

#include "ExtendedProblemStat.h"
#include "Refinement.h"
#include <boost/lexical_cast.hpp>

namespace AMDiS {

/// Implementation of ProblemStat to allow for the conservation of mass of one
/// solution component, i.e. project the old-solution to the new mesh so that
/// the scalar-product is conserved: \f$ (u^*, v) = (u^\text{old}, v) \f$, where \f$ u^*,v \f$ live on the
/// new mesh and \f$ u^\text{old} \f$ lives on the old mesh.
struct ProblemStatMassConserve : public ExtendedProblemStat, public StandardRefineOperation
{
  ProblemStatMassConserve(std::string name_)
      : ExtendedProblemStat(name_),
	initialMeshAdoption(false),
	comp(-1),
	old_solution(NULL)
  {
    Parameters::get(name_ + "->mass conservation component", comp);
    assert( comp >= 0 );
    
    Parameters::get(name_ + "->additional components", nAddComponents);
    nAddComponents++;
    Parameters::set(name_ + "->additional components", nAddComponents);
    
    comp_old = nComponents - 1 + nAddComponents;
    
    int degree = 1;
    Parameters::get(name_ + "->polynomial degree[" + boost::lexical_cast<std::string>(comp) + "]", degree);
    Parameters::set(name_ + "->polynomial degree[" + boost::lexical_cast<std::string>(comp_old) + "]", degree);
    
    int refSet = 123;
    Parameters::set(name_ + "->refinement set[" + boost::lexical_cast<std::string>(comp_old) + "]", refSet);
  }
  
  ~ProblemStatMassConserve()
  {
    if (old_solution) {
      delete old_solution; old_solution = NULL;
    }
  }
  
  virtual void initialize(Flag initFlag,
			  ProblemStatSeq *adoptProblem = NULL,
			  Flag adoptFlag = INIT_NOTHING) override
  {
    ExtendedProblemStat::initialize(initFlag, adoptProblem, adoptFlag);
    old_solution = new DOFVector<double>(getFeSpace(comp_old), "old-solution");
  }
  
  void buildAfterCoarsen(AdaptInfo *adaptInfo, Flag flag,
				bool assembleMatrix = true,
				bool assembleVector = true) override
  { FUNCNAME("ProblemStatMassConserve_mm::buildAfterCoarsen()");
  
    if (!initialMeshAdoption) {
      fitMeshes(comp, comp_old);
      old_solution->interpol(getSolution(comp));
      initialMeshAdoption = true;
    }
    ExtendedProblemStat::buildAfterCoarsen(adaptInfo, flag, assembleMatrix, assembleVector);
  }
  
  /// copy solution to old_solution after regular solve
  void solve(AdaptInfo *adaptInfo, 
	     bool createMatrixData = true, 
	     bool storeMatrixData = false) override
  { FUNCNAME("ProblemStatMassConserve_mm::solve()");
  
    ExtendedProblemStat::solve(adaptInfo, createMatrixData, storeMatrixData);
    
    fitMeshes(comp, comp_old);
    old_solution->interpol(getSolution(comp));
  }
  
  DOFVector<double>* getOldSolution(int comp0 = 0) { assert( comp0 == comp ); return old_solution; }
  
protected:
  void fitMeshes(int comp1, int comp2)
  {
    MeshStructure meshStructure2;
    meshStructure2.init(getMesh(comp1));
    getCoarseningManager()->globalCoarsen(getMesh(comp2), -2);
    meshStructure2.fitMeshToStructure(getMesh(comp2), getRefinementManager());
  }
  
private:
  /// true if initial-solution/-mesh is copied to temporary problem
  bool initialMeshAdoption; 	
  
  /// component the should be copnserved
  int comp;
  int comp_old;
  
  DOFVector<double>* old_solution;
};

}

#endif // PROBLEM_STAT_MASS_CONSERVE