/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_REFINEMENT_LEVEL_PHASEFIELD_H
#define EXTENSIONS_REFINEMENT_LEVEL_PHASEFIELD_H

#include "ElementFunction.h"
#include "Refinement.h"

namespace AMDiS { namespace extensions {


class RefinementLevelSimple : public RefinementLevel<WorldVector<double>, int >
{
public:
  RefinementLevelSimple(const FiniteElemSpace *feSpace_, int level_) :
    RefinementLevel<WorldVector<double>, int >(feSpace_), refineLevel(level_) { }
    
  ~RefinementLevelSimple() {}

  void markElements(Flag &markFlag) 
  {
    bool elMarkRefine = false, elMarkCoarsen = false;

    Flag traverseFlag = Mesh::CALL_LEAF_EL;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    while (elInfo) {
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }
    
    markFlag = 0;
    if (elMarkRefine) 
      markFlag = 1;
    if (elMarkCoarsen) 
      markFlag |= 2;
  }
  
protected:
  int refineLevel;
};

/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to given
 * coordinates.
 **/
class RefinementLevelCoords : public RefinementLevel<WorldVector<double>, int >
{
public:
  RefinementLevelCoords(const FiniteElemSpace *feSpace_, MeshRefinementFunction<WorldVector<double>, int >* refineFct_) :
    RefinementLevel<WorldVector<double>, int >(feSpace_, refineFct_) { }
    
  ~RefinementLevelCoords() {}

  void markElements(Flag &markFlag) 
  {
    FixVec<WorldVector<double>, VERTEX> coords(mesh->getDim(), NO_INIT);
    WorldVector<double> center;

    bool elMarkRefine = false, elMarkCoarsen = false;

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    while (elInfo) {
      coords = elInfo->getCoords();
      center.set(0.0);
      for (int i = 0; i < coords.getSize(); ++i) {
	center += coords[i];
      }
      center *= 1.0 / coords.getSize();
      int refineLevel = (*refineFct)(center);
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }
    
    markFlag = 0;
    if (elMarkRefine) 
      markFlag = 1;
    if (elMarkCoarsen) 
      markFlag |= 2;
  }
};


/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to given
 * coordinates.
 **/
class RefinementLevelCoords2 : public RefinementLevel<double, int >
{
public:
  RefinementLevelCoords2(const FiniteElemSpace *feSpace_, MeshRefinementFunction<double, int >* refineFct_, AbstractFunction<double, WorldVector<double> > *fct_) :
    RefinementLevel<double, int >(feSpace_, refineFct_), fct(fct_) { }
    
  ~RefinementLevelCoords2() {}

  void markElements(Flag &markFlag)
  {
    FixVec<WorldVector<double>, VERTEX> coords(mesh->getDim(), NO_INIT);
    WorldVector<double> center;

    bool elMarkRefine = false, elMarkCoarsen = false;

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    while (elInfo) {
      coords = elInfo->getCoords();
      center.set(0.0);
      for (int i = 0; i < coords.getSize(); ++i) {
	center += coords[i];
      }
      center *= 1.0 / coords.getSize();
      double value = (*fct)(center);
      int refineLevel = (*refineFct)(value);
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }

    markFlag = 0;
    if (elMarkRefine)
      markFlag = 1;
    if (elMarkCoarsen)
      markFlag |= 2;
  }

private:
  AbstractFunction<double, WorldVector<double> > *fct;
};


/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to a given
 * value, like phase-value or distance-value.
 **/
class RefinementLevelDOF : public RefinementLevel<double, int>
{
public:
  RefinementLevelDOF(const FiniteElemSpace *feSpace_, MeshRefinementFunction<double, int>* refineFct_, DOFVector<double> *vec_) :
    RefinementLevel<double, int>(feSpace_, refineFct_), 
    vec(vec_) { }
    
  ~RefinementLevelDOF() {}

  void markElements(Flag &markFlag) 
  {
    double meanValue;

    ElementFunctionDOFVec<double> elFct(vec);
    DimVec<double> lambda(mesh->getDim(), NO_INIT);
    for (int i = 0; i < lambda.getSize(); ++i)
      lambda[i] = 1.0 / lambda.getSize();

    bool elMarkRefine = false, elMarkCoarsen = false;

    Flag traverseFlag = Mesh::CALL_LEAF_EL;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    while (elInfo) {
      elFct.setElInfo(elInfo);
      meanValue = elFct(lambda);

      int refineLevel = (*refineFct)(meanValue);
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }
		
    markFlag = 0;
    if (elMarkRefine) 
      markFlag = 1;
    if (elMarkCoarsen) 
      markFlag |= 2;
  }

private:

  DOFVector<double>* vec;
};


/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to a given
 * value, like phase-value or distance-value, and the center coords of the
 * element.
 **/
class RefinementLevelCoordsDOF : public RefinementLevel<std::pair<WorldVector<double>, double>, int >
{
public:

  RefinementLevelCoordsDOF(const FiniteElemSpace *feSpace_, MeshRefinementFunction<std::pair<WorldVector<double>, double>, int >* refineFct_, DOFVector<double> *vec_) :
    RefinementLevel<std::pair<WorldVector<double>, double>, int >(feSpace_, refineFct_), vec(vec_) { }
    
  ~RefinementLevelCoordsDOF() {}

  void markElements(Flag &markFlag) 
  {
    double meanValue;
    FixVec<WorldVector<double>, VERTEX> coords(mesh->getDim(), NO_INIT);
    WorldVector<double> center;

    ElementFunctionDOFVec<double> elFct(vec);
    DimVec<double> lambda(mesh->getDim(), NO_INIT);
    for (int i= 0; i<lambda.getSize(); ++i)
      lambda[i]= 1.0/lambda.getSize();

    bool elMarkRefine= false, elMarkCoarsen= false;

    Flag traverseFlag= Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS;
    TraverseStack stack;
    ElInfo *elInfo= stack.traverseFirst(mesh, -1, traverseFlag);
    while (elInfo) {
      elFct.setElInfo(elInfo);
      meanValue = elFct(lambda);
      coords = elInfo->getCoords();
      center.set(0.0);
      for (int i = 0; i < coords.getSize(); ++i) {
        center += coords[i];
      }
      center *= 1.0 / coords.getSize();
      int refineLevel = (*refineFct)(std::make_pair(center, meanValue));
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }
                
    markFlag = 0;
    if (elMarkRefine) 
      markFlag = 1;
    if (elMarkCoarsen) 
      markFlag |= 2;
  }

private:

  DOFVector<double>* vec;
};


/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to a given
 * list of values, like phase-values or distance-values.
 **/
class RefinementLevelDOFList : public RefinementLevel< std::vector<double>, int >
{
public:
  RefinementLevelDOFList(
      const FiniteElemSpace *feSpace_, 
      MeshRefinementFunction< std::vector<double>, int >* refineFct_, 
      DOFVector<double>* vec1_, 
      DOFVector<double>* vec2_ = NULL,
      DOFVector<double>* vec3_ = NULL) : 
    RefinementLevel< std::vector<double>, int >(feSpace_, refineFct_)
  {
    vecs.push_back(vec1_);
    if(vec2_ != NULL) vecs.push_back(vec2_);
    if(vec3_ != NULL) vecs.push_back(vec3_);			
  }
    
  ~RefinementLevelDOFList() {}

  void markElements(Flag &markFlag) 
  {    
    bool elMarkRefine = false, elMarkCoarsen = false;

    std::vector<ElementFunctionDOFVec<double>*> elFcts;
    for (unsigned i = 0; i < vecs.size(); ++i) {
      elFcts.push_back(new ElementFunctionDOFVec<double>(vecs[i]));
    }
    DimVec<double> lambda(mesh->getDim(), NO_INIT);
    for (int i= 0; i < lambda.getSize(); ++i)
      lambda[i] = 1.0 / lambda.getSize();

      Flag traverseFlag = Mesh::CALL_LEAF_EL;
      TraverseStack stack;
      ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
      while(elInfo) {
	std::vector<double> meanValues;
	for (unsigned v = 0; v < vecs.size(); ++v) {
          elFcts[v]->setElInfo(elInfo);
	  meanValues.push_back( (*(elFcts[v]))(lambda) );
	}
	
        int refineLevel = (*refineFct)(meanValues);
        int oldLevel = elInfo->getLevel();
        elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

	elMarkRefine |= elInfo->getElement()->getMark() == 1;
	elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
	elInfo = stack.traverseNext(elInfo);
    }
      
    markFlag= 0;
    if(elMarkRefine) 
	    markFlag= 1;
    if(elMarkCoarsen) 
	    markFlag|= 2;
  }
  
private:
	std::vector<DOFVector<double>*> vecs;
};



/** \brief
 * Refinement structure to perform local anisotropic refinement depending
 * on a refineFunction that defines the local refinement level to a given
 * list of values, like phase-values or distance-values.
 **/
class RefinementLevelES : public RefinementLevel< std::vector<double>, int >
{
public:
  RefinementLevelES(
      const FiniteElemSpace *feSpace_, 
      MeshRefinementFunction< std::vector<double>, int >* refineFct_, 
      DOFVector<double>* vec1_, 
      DOFVector<double>* vec2_, 
      DOFVector<double>* phase_,
      double tol_) : 
    RefinementLevel< std::vector<double>, int >(feSpace_, refineFct_),
    phase(phase_),
    tol(tol_)
  {
    vecs.push_back(vec1_);
    vecs.push_back(vec2_);
  }
    
  ~RefinementLevelES() {}

  void markElements(Flag &markFlag) 
  {
    FUNCNAME("RefinementLevelDOF::markElements()");

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int numBasFcts = basFcts->getNumber();

    DegreeOfFreedom *localIndices = new DegreeOfFreedom[numBasFcts];
    FixVec<WorldVector<double>, VERTEX> coords(mesh->getDim(), NO_INIT);

    bool elMarkRefine = false, elMarkCoarsen = false;

    std::vector<ElementFunctionDOFVec<double>*> elFcts;
    for (unsigned i = 0; i < vecs.size(); ++i) {
      elFcts.push_back(new ElementFunctionDOFVec<double>(vecs[i]));
    }
    ElementFunctionDOFVec<double>* elFct_phase = new ElementFunctionDOFVec<double>(phase);
    DimVec<double> lambda(mesh->getDim(), NO_INIT);
    for (int i= 0; i < lambda.getSize(); ++i)
      lambda[i] = 1.0 / lambda.getSize();

    Flag traverseFlag = Mesh::CALL_LEAF_EL;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);

//     int n = 0;
//     double mean = 0.0;
//     while(elInfo) {
//       elFct_phase->setElInfo(elInfo);
//       if ((*elFct_phase)(lambda) > 0.001) {
//         std::vector<double> meanValues;
//         meanValues.push_back(0.0);
//         for (unsigned v = 0; v < vecs.size(); ++v) {
//           elFcts[v]->setElInfo(elInfo);
//           meanValues.push_back( (*(elFcts[v]))(lambda) );
//         }
//         
//         double value = refineFct->indicator(meanValues);
//         mean += value;
//         n++;
//       }
//       elInfo = stack.traverseNext(elInfo);
//     }
//     mean /= n;
//     MSG("mean value = %e\n", mean);

    elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    int nCoarse = 0;
    while(elInfo) {
      elFct_phase->setElInfo(elInfo);
      if ((*elFct_phase)(lambda) > 0.0001) {
        std::vector<double> meanValues;
        meanValues.push_back(tol);
        for (unsigned v = 0; v < vecs.size(); ++v) {
          elFcts[v]->setElInfo(elInfo);
          meanValues.push_back( (*(elFcts[v]))(lambda) );
        }
        
        int refineLevel = (*refineFct)(meanValues);
        nCoarse += refineLevel;
        elInfo->getElement()->setMark( refineLevel );
      } else {
        elInfo->getElement()->setMark( -1 );
      }

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }
    MSG("number of elements marked for coarsening: %d\n", abs(nCoarse));
      
    markFlag= 0;
    if(elMarkRefine) 
            markFlag= 1;
    if(elMarkCoarsen) 
            markFlag|= 2;
    
    delete [] localIndices;
  }
  
private:
        std::vector<DOFVector<double>*> vecs;
        DOFVector<double>* phase;
        double tol;
};

} }

#endif // EXTENSIONS_REFINEMENT_LEVEL_PHASEFIELD_H

