/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "NavierStokesCahnHilliard.h"
// #include "Views.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"
 
namespace AMDiS { namespace base_problems {

NavierStokesCahnHilliard::NavierStokesCahnHilliard(const std::string &name_, bool createProblem) :
  super(name_, createProblem),
  viscosity1(1.0),
  viscosity2(1.0),
  density1(1.0),
  density2(1.0),    
  densityPhase(NULL),
  viscosityPhase(NULL),
  multiPhase(NULL),
  velocity(NULL),
  fileWriter(NULL),
  useMobility(false),
  refFunction(NULL),
  refinement(NULL),
  sigma(0.0),
  surfaceTension(0.0),
  doubleWell(0),
  gamma(1.0),
  eps(0.1),
  minusEps(-0.1),
  epsInv(10.0),
  minusEpsInv(-10.0),
  epsSqr(0.01),
  minusEpsSqr(-0.01)
  {
    dow = Global::getGeo(WORLD);
  Initfile::get(name + "->viscosity1", viscosity1); // viscosity of fluid 1
  Initfile::get(name + "->viscosity2", viscosity2); // viscosity of fluid 2
  Initfile::get(name + "->density1", density1); // density of fluid 1
  Initfile::get(name + "->density2", density2); // density of fluid 2

  Initfile::get(name + "->non-linear term", nonLinTerm); 
  Initfile::get(name + "->theta", theta); 
  Initfile::get("adapt->timestep", tau); 
    
    
    Parameters::get(name + "->epsilon", eps); // interface width
    Parameters::get(name + "->sigma", sigma); 
    double a_fac;
    Parameters::get(name + "->a_factor", a_fac);
    a=1.0/sigma*sqr(eps)*a_fac* (std::max(density1,density2)/tau);
    b=1.0;
    ab=a*b;   
    surfaceTension = -sigma/eps * 3.0/2.0/sqrt(2.0) * a ;

  theta1 = 1.0-theta;
  minusTheta1 = -theta1;

  force.set(0.0);
  Initfile::get(name + "->force", force);
  
  // parameters for CH
  Parameters::get(name + "->use mobility", useMobility); // mobility
  Parameters::get(name + "->gamma", gamma); // mobility

  
  // type of double well: 0= [0,1], 1= [-1,1]
  Parameters::get(name + "->double-well type", doubleWell); 
  
  // transformation of the parameters
  minusEps = -eps;
  minus1 = -1.0;
  epsInv = 1.0/eps;
  minusEpsInv = -epsInv;
  epsSqr = sqr(eps);
  minusEpsSqr = -epsSqr*b;
  
  
}


NavierStokesCahnHilliard::~NavierStokesCahnHilliard() 
{ 
  delete densityPhase;
  delete viscosityPhase;
}


void NavierStokesCahnHilliard::setTime(AdaptInfo* adaptInfo) 
{
  super::setTime(adaptInfo);
  delta = gamma*adaptInfo->getTimestep();
}


void NavierStokesCahnHilliard::initData()
{ 
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    string initFileStr = name + "->space->solver", solverType = "";
    Parameters::get(initFileStr, solverType);
    if (solverType == "petsc-nsch") {
      Parallel::PetscSolverNSCH* solver = dynamic_cast<Parallel::PetscSolverNSCH*>(prob->getSolver()); 
      if (solver)
      {
	solver->setChData(&eps, &delta);
	solver->setStokesData( getInvTau(), prob->getSolution(), &viscosity1, &viscosity2, &density1, &density2);
      }
    }
#endif

  densityPhase = new DOFVector<double>(prob->getFeSpace(0), "densityPhase");
  viscosityPhase = new DOFVector<double>(prob->getFeSpace(0), "viscosityPhase");

  densityPhase->set(density1);
  viscosityPhase->set(viscosity1);
  
   if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(getFeSpace(0), "velocity");
 
  fileWriter = new FileVectorWriter(name + "->velocity->output", getFeSpace()->getMesh(), velocity);

  super::initData();
}


void NavierStokesCahnHilliard::closeTimestep(AdaptInfo *adaptInfo)
{ 
  super::closeTimestep(adaptInfo);
  calcVelocity();
  fileWriter->writeFiles(adaptInfo, true);
}


void NavierStokesCahnHilliard::transferInitialSolution(AdaptInfo *adaptInfo)
{ 
  calcVelocity();

  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);

  // initial parameters for detecting mesh changes
  oldMeshChangeIdx= getMesh()->getChangeIndex();
}


void NavierStokesCahnHilliard::initTimestep(AdaptInfo *adaptInfo)
{ 
  super::initTimestep(adaptInfo);
  refinement->refine(2);
  LinearInterpolation1 dLI(density1, density2);
  LinearInterpolation1 vLI(viscosity1, viscosity2);
  transformDOF(prob->getSolution()->getDOFVector(dow+2), densityPhase, &dLI);
  transformDOF(prob->getSolution()->getDOFVector(dow+2), viscosityPhase, &vLI);
}



void NavierStokesCahnHilliard::solveInitialProblem(AdaptInfo *adaptInfo)
{
  using namespace extensions;
  // meshFunction for klocal refinement around the interface of the phasefield-DOFVector
  refFunction = new PhaseFieldRefinement(prob->getMesh());

  if (getDoubleWellType() == 0) {
    refinement = new RefinementLevelDOF(
      prob->getFeSpace(),
      refFunction,
      prob->getSolution()->getDOFVector(dow+2)); // phaseField-DOFVector in [0,1]
  } else {
    refinement = new RefinementLevelDOF(
      prob->getFeSpace(),
      refFunction,
      new PhaseDOFView<double>(prob->getSolution()->getDOFVector(dow+2))); // phaseField-DOFVector in [-1,1]
  }

  // initial refinement
  refinement->refine(0);

  for (int i = 0; i < 3; i++) {
    solveInitialProblem2(adaptInfo); 	// update phaseField-DOFVector
    refinement->refine((i < 4 ? 4 : 10));	// do k steps of local refinement
  }

  // solve all initial problems of the problems added to the CouplingTimeInterface    
}



void NavierStokesCahnHilliard::solveInitialProblem2(AdaptInfo *adaptInfo) 
{ 
  using namespace extensions;
  Flag initFlag = initDataFromFile(adaptInfo);
  
  int comp = 2+dow;

  if (!initFlag.isSet(DATA_ADOPTED)) {
    int initialInterface = 0;
    Initfile::get(name + "->initial interface", initialInterface);
    double initialEps = eps;
    Initfile::get(name + "->initial epsilon", initialEps);

    if (initialInterface == 0) {
      /// horizontale Linie
      double a= 0.0, dir= -1.0;
      Initfile::get(name + "->line->pos", a);
      Initfile::get(name + "->line->direction", dir);
      prob->getSolution()->getDOFVector(comp)->interpol(new Plane(a, dir));
    }
    else if (initialInterface == 1) {
      /// schraege Linie
      double theta = m_pi/4.0;
      prob->getSolution()->getDOFVector(comp)->interpol(new PlaneRotation(0.0, theta, 1.0));
      transformDOFInterpolation(prob->getSolution()->getDOFVector(comp),new PlaneRotation(0.0, -theta, -1.0), new AMDiS::Min<double>);
    }
    else if (initialInterface == 2) {
      /// Ellipse
      double a= 1.0, b= 1.0;
      Initfile::get(name + "->ellipse->a", a);
      Initfile::get(name + "->ellipse->b", b);
      prob->getSolution()->getDOFVector(comp)->interpol(new Ellipse(a,b));
    }
    else if (initialInterface == 3) {
      /// zwei horizontale Linien
      double a= 0.75, b= 0.375;
      Initfile::get(name + "->lines->pos1", a);
      Initfile::get(name + "->lines->pos2", b);
      prob->getSolution()->getDOFVector(comp)->interpol(new Plane(a, -1.0));
      transformDOFInterpolation(prob->getSolution()->getDOFVector(comp),new Plane(b, 1.0), new AMDiS::Max<double>);
    }
    else if (initialInterface == 4) {
      /// Kreis
      double radius= 1.0, x=0, y=0;
      Initfile::get(name + "->circle->radius", radius);
      Initfile::get(name + "->circle->center_x", x);
      Initfile::get(name + "->circle->center_y", y);
      WorldVector<double> w;
      w[0]=x; w[1]=y+adaptInfo->getTime();
      prob->getSolution()->getDOFVector(comp)->interpol(new Circle(radius, w));
    } else if (initialInterface == 5) {
      /// Rechteck
      double width = 0.5;
      double height = 0.3;
      WorldVector<double> center; center.set(0.5);
      Initfile::get(name + "->rectangle->width", width);
      Initfile::get(name + "->rectangle->height", height);
      Initfile::get(name + "->rectangle->center", center);
      prob->getSolution()->getDOFVector(comp)->interpol(new Rectangle(width, height, center));
    }

    /// create phase-field from signed-dist-function
    if (doubleWell == 0) {
      transformDOF(prob->getSolution()->getDOFVector(comp),
        new SignedDistToPhaseField(initialEps));
    } else {
      transformDOF(prob->getSolution()->getDOFVector(comp),
        new SignedDistToCh(initialEps));
    }
  }
}


void NavierStokesCahnHilliard::fillOperators()
{ 
  using namespace extensions;
  const FiniteElemSpace* feSpace = prob->getFeSpace(0);
  int comp = dow+2, comp2 = dow+1;
  int degree = prob->getFeSpace(comp)->getBasisFcts()->getDegree();
 
  // c
  Operator *opChMnew = new Operator(feSpace,feSpace);
  opChMnew->addZeroOrderTerm(new Simple_ZOT);
  Operator *opChMold = new Operator(feSpace,feSpace);
  opChMold->addZeroOrderTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(comp)));
  // -nabla*(grad(c))
  Operator *opChL = new Operator(feSpace,feSpace);
  opChL->addSecondOrderTerm(new Simple_SOT);
  
  // div(M(c)grad(mu)), with M(c)=gamma/4*(c^2-1)^2
  Operator *opChLM = new Operator(feSpace,feSpace);
  opChLM->addSecondOrderTerm(new Simple_SOT(gamma*ab));
  
  // -2*c_old^3 + 3/2*c_old^2
  Operator *opChMPowExpl = new Operator(feSpace,feSpace);
  opChMPowExpl->addZeroOrderTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(comp),
    new AMDiS::Pow<3>(-2.0, 3*degree)));
  if (doubleWell == 0) {
    opChMPowExpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getSolution()->getDOFVector(comp),
      new AMDiS::Pow<2>(3.0/2.0, 2*degree)));
  }

  // -3*c_old^2 * c
  Operator *opChMPowImpl = new Operator(feSpace,feSpace);
  opChMPowImpl->addZeroOrderTerm(new VecAtQP_ZOT(
    prob->getSolution()->getDOFVector(comp),
    new AMDiS::Pow<2>(-3.0, 2*degree)));
  if (doubleWell == 0) {
    opChMPowImpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getSolution()->getDOFVector(comp),
      new AMDiS::Factor<double>(3.0, degree)));
    opChMPowImpl->addZeroOrderTerm(new Simple_ZOT(-0.5));
  } else {
    opChMPowImpl->addZeroOrderTerm(new Simple_ZOT(1.0));
  }

  // mu + eps^2*laplace(c) + c - 3*(c_old^2)*c = -2*c_old^3 [+ BC]
  // ----------------------------------------------------------------------
  prob->addMatrixOperator(*opChMnew,comp2,comp2, &ab);              /// < phi*mu , psi >
  
  prob->addMatrixOperator(*opChMPowImpl,comp2,comp, &b);          /// < -3*phi*c*c_old^2 , psi >
  prob->addMatrixOperator(*opChL,comp2,comp, &minusEpsSqr);   /// < -eps^2*phi*grad(c) , grad(psi) >
  // . . . vectorOperators . . . . . . . . . . . . . . .
  prob->addVectorOperator(*opChMPowExpl,comp2, &b);            /// < -2*phi*c_old^3 , psi >

//   setAssembleMatrixOnlyOnce_butTimestepChange(0,1);
  
  // dt(c) = laplace(mu) - u*grad(c)
  // -----------------------------------
  prob->addMatrixOperator(*opChMnew,comp,comp, &b); /// < phi*c/tau , psi >
  prob->addMatrixOperator(*opChLM,comp,comp2, getTau());                /// < phi*grad(mu) , grad(psi) >
  // . . . vectorOperators . . . . . . . . . . . . . . .
  prob->addVectorOperator(*opChMold,comp, &b );   /// < phi*c^old/tau , psi >
  
  /**/
  
  
  /// Navier-Stokes part
  int degree_u = getFeSpace(0)->getBasisFcts()->getDegree();
//   int degree_p = getFeSpace(dow)->getBasisFcts()->getDegree();
    
  WorldVector<DOFVector<double>* > vel;
  for (unsigned i = 0; i < dow; i++)
    vel[i]= prob->getSolution()->getDOFVector(i);


  // fill operators for prob
  for (unsigned i = 0; i < dow; ++i) {
    
    /// < (1/tau)*rho*u'_i , psi >
    Operator *opTime = new Operator(getFeSpace(i), getFeSpace(i));
    if (density1==density2)
      opTime->addTerm(new Simple_ZOT(density1));  
    else
      opTime->addTerm(new VecAtQP_ZOT(densityPhase));
    opTime->setUhOld(prob->getSolution()->getDOFVector(i));
    prob->addMatrixOperator(*opTime, i, i, getInvTau(), getInvTau());    
    prob->addVectorOperator(*opTime, i, getInvTau(), getInvTau());    
    
    
 
    /// < u^old*grad(u_i^old) , psi >
/*    Operator *opUGradU0 = new Operator(getFeSpace(i),getFeSpace(i));
    if (density1==density2)
        opUGradU0->addTerm(new WorldVec_FOT(vel, -1.0), GRD_PHI);
    else
        opUGradU0->addTerm(new WorldVecPhase_FOT(densityPhase, vel, -1.0), GRD_PHI);
    opUGradU0->setUhOld(prob->getSolution()->getDOFVector(i));
    if (nonLinTerm == 0) {
      prob->addVectorOperator(*opUGradU0, i);
    } else {
      prob->addVectorOperator(*opUGradU0, i, &theta1, &theta1);
    }

    if (nonLinTerm == 1) {
      /// < u'*grad(u_i^old) , psi >
      for (unsigned j = 2; j < 2+dow; ++j) {
        Operator *opUGradU1 = new Operator(getFeSpace(i),getFeSpace(i));
	if (density1==density2)
	  opUGradU1->addTerm(new PartialDerivative_ZOT(prob->getSolution()->getDOFVector(i), j-2));
	else
	  opUGradU1->addTerm(new VecAndPartialDerivative_ZOT(     densityPhase,    prob->getSolution()->getDOFVector(i), j-2));
        prob->addMatrixOperator(*opUGradU1, i, j, &theta, &theta);
      }
    } else if (nonLinTerm == 2) {
      /// < u^old*grad(u'_i) , psi >
*/   for(unsigned j = 2; j < 2+dow; ++j) {
        Operator *opUGradU2 = new Operator(getFeSpace(i),getFeSpace(i));
        opUGradU2->addTerm(new Vec2AndPartialDerivative_FOT(   densityPhase,    prob->getSolution()->getDOFVector(j-2), j-2), GRD_PHI);
        prob->addMatrixOperator(*opUGradU2, i, i);
      }
    
  /**/
    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    Operator *opLaplaceUi = new Operator(getFeSpace(i), getFeSpace(i));
    opLaplaceUi->addTerm(new VecAtQP_SOT(viscosityPhase));
    prob->addMatrixOperator(*opLaplaceUi, i, i);  
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(i),getFeSpace(dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    prob->addMatrixOperator(*opGradP, i, dow);
    
    /// external force, i.e. gravitational force
    if (force[i]*force[i] > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      opForce->addZeroOrderTerm(new VecAtQP_ZOT(densityPhase, new Force(force[i], degree_u)));
      prob->addVectorOperator(*opForce, i);
    }
  
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(dow),getFeSpace(i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, dow, i);
    
    
    
     ///coupling Operators 
      Operator *opNuGradC = new Operator(getFeSpace(i), getFeSpace(comp2));
      opNuGradC->addTerm(new PartialDerivative_ZOT(prob->getSolution()->getDOFVector(comp), i));
      prob->addMatrixOperator(opNuGradC, i, comp2, &surfaceTension, &surfaceTension);
    
      Operator *opVGradC = new Operator(getFeSpace(comp), getFeSpace(i));
      opVGradC->addTerm(new PartialDerivative_ZOT(prob->getSolution()->getDOFVector(comp), i, b));
      prob->addMatrixOperator(opVGradC, comp, i,  getTau());
      /**/
     
  }
  
  /**/
}


void NavierStokesCahnHilliard::addLaplaceTerm(int i)
{ 
  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (viscosity1!=viscosity2) {
    for (unsigned j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(i), getFeSpace(j));
      opLaplaceUi1->addTerm(new VecAtQP_IJ_SOT(  viscosityPhase, NULL, j, i));      
      prob->addMatrixOperator(*opLaplaceUi1, i, j);      
    }
  }/**/
  
  /// < alpha*grad(u'_i) , grad(psi) >
  
 
}

} }
