/** \file PhaseFieldCrystal.h */

#ifndef PHASE_FIELD_CRYSTAL_H
#define PHASE_FIELD_CRYSTAL_H

#include "AMDiS.h"
#include "BaseProblem.h"
#include "ExtendedProblemStat.h"

namespace AMDiS { namespace base_problems {

namespace detail {

  /** Phase-field Crystal problem:
  *
  *       mu = psi^3 + (1+r)*psi + 2*laplace(psi) + laplace(nu)
  * d_t(psi) = div( (psi + 1.5) * grad(mu) )
  *       nu = laplace(psi)
  */
  template<typename ProblemStatType>
  class PhaseFieldCrystal : public BaseProblem<ProblemStatType>
  {
  public: // typedefs

    typedef PhaseFieldCrystal<ProblemStatType> self;
    typedef BaseProblem<ProblemStatType> super;

  public:
  
    PhaseFieldCrystal(const std::string &name_, bool createProblem = true);
    
    ~PhaseFieldCrystal()
    {
      if (oldSolution) {
	delete oldSolution;
	oldSolution = NULL;
      }
      if (stepSolution) {
	delete stepSolution;
	stepSolution = NULL;
      }
    }

    DOFVector<double>* getDensity() { return self::prob->getSolution(1); }
    double calcEnergy();

    virtual void initData() override ;
    virtual void transferInitialSolution(AdaptInfo *adaptInfo) override;
    
    virtual void closeTimestep(AdaptInfo *adaptInfo) override;
    
    virtual void fillOperators() override ;
    virtual void fillBoundaryConditions() override {};
    
    void rejectTimestep()
    {
      *stepSolution = *oldSolution;
    }
    
    void acceptTimestep()
    { FUNCNAME("PhaseFieldCrystal::acceptTimestep");
      MSG("Timestep accepted.\n");
      *oldSolution = *self::prob->getSolution();
      *stepSolution = *self::prob->getSolution();
    }
    
    SystemVector* getStepSolution()
    {
      return stepSolution;
    }
    
    SystemVector* getOldSolution()
    {
      return oldSolution;
    }
    
  protected:

    int mobility_type;

    double r;
    double density;
    double M0;
    
    SystemVector* oldSolution;
    SystemVector* stepSolution;
  };

} // end namespace detail

typedef detail::PhaseFieldCrystal<AMDiS::ProblemStat> PhaseFieldCrystal;


} }

#include "PhaseFieldCrystal.hh"

#endif // PHASE_FIELD_CRYSTAL_H
