/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef PHASE_FIELD_CRYSTAL_H
#define PHASE_FIELD_CRYSTAL_H

#include "AMDiS.h"
#include "BaseProblem_RB.h"


namespace AMDiS { namespace base_problems {

/** Phase-field Crystal problem
 */
class PhaseFieldCrystal_RB : public BaseProblem_RB
{
public: // typedefs

  typedef BaseProblem_RB super;

public:

  PhaseFieldCrystal_RB(const std::string &name_, bool createProblem = true);
  ~PhaseFieldCrystal_RB() {};

  void fillOperators() override;
  void fillBoundaryConditions() override;

protected:

  bool useMobility;

  double tempParameter;
  double r;
  double rho0;
  double density;
  double two;
  double minus2;
  double minus1;
};


/** \ingroup MainInstat
 * \brief
 * Abstract function to calculate the pure PFC-Energy
 */
class Energy : public BinaryAbstractFunction<double,double,double>
{
  public:
    Energy() : BinaryAbstractFunction<double,double,double>(4) { }

    double operator()(const double &rho, const double &mu) const {
      return -0.25*sqr(sqr(rho)) + 0.5*rho*mu; }
};


class MobilityPfc : public AbstractFunction<double,double>
{
  public:
    MobilityPfc(double density_ = -0.3, double factor_ = 1.0, int degree=1) : 
      AbstractFunction<double,double>(degree+1),
      density(density_),
      factor(factor_),
      delta(1.e-2) {
	rho0 = 3.0*density;
      }

    double operator()(const double &rho) const 
    {
      double mobility= (rho - rho0)*factor;
      return std::max(delta, mobility);
    }
  
  protected:
    double density;
    double factor;
    double delta;
    
    double rho0;
};


class MobilityPfcDiff : public AbstractFunction<double,double>
{
  public:
    MobilityPfcDiff(double density_ = -0.3, double factor_ = 1.0, int degree=1) : 
      AbstractFunction<double,double>(2), // höherer Quadraturgrad, weil Sprung aufgelöst werden muss
      density(density_),
      factor(factor_),
      delta(1.e-2) {
	rho0 = 3.0*density;	
      }

    double operator()(const double &rho) const 
    {
      double mobility= (rho > rho0+delta ? factor : delta);
      return mobility;
    }
  
  protected:
    double density;
    double factor;
    double delta;
    
    double rho0;
};

} }

#endif // PHASE_FIELD_CRYSTAL_H
