/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef POLARIZATION_FIELD_RB_H
#define POLARIZATION_FIELD_RB_H

#include "AMDiS.h"
#include "BaseProblem_RB.h"
#include "GenericOperatorTerm.h"


namespace AMDiS { namespace base_problems {

namespace detail 
{
  
  /** \ingroup BaseProblems
  * \brief
  * Simulation of the relaxation of an orientation field (polarization field).
  * The evolution equation reads:
  * 
  * d_t P = alpha2*laplace(P#) - alpha4*P#
  *     P#= 1/eps * (|P|^2 - 1)*P - K*laplace(P)
  * 
  * with P=(P1, P2), P#=(P#1, P#2) and 
  * |P| = sqrt(P^T*P), [laplace(P)]_i = laplace(P_i)
  * 
  * As time-discretization a Rosenbrock scheme is implemented. You have to use
  * an ExtendedRosenbrockAdaptInstationary<ThisType> to run the time evolution.
  * Therefor you have to set Rosenbrock parameters in the initfile:
  * "[name]->space->rosenbrock->method: METHOD"
  * "[name]->space->rosenbrock->error weights: [1,1,0,0]"
  * "adapt[0]->time tolerance: TOL"
  * 
  * A filewriter for P is provided via the initfile parameter
  * "[name]->vectorField->output". You have to set the sup-parameter 
  * "(...)->ParaView vector format: 1" and 
  * "(...)->write vector as 3d vector: 1" to visualize the field P as 
  * vectorfield in ParaView.
  */  
  template<typename ProblemStatType>
  class PolarizationField_RB : public BaseProblem_RB //<ProblemStatType>
  {
  public: // typedefs
    
    typedef PolarizationField_RB<ProblemStatType> self;
    typedef BaseProblem_RB/*<ProblemStatType>*/ super;
    
  public: // methods

    /// constructor
    PolarizationField_RB(const std::string &name_);

    /// destructor
    ~PolarizationField_RB();

    /// initialize the vectorField and corresponding fileWriter
    void initData() override;

    /// calls \ref calcVectorField and \ref super::transferInitialSolution
    void transferInitialSolution(AdaptInfo *adaptInfo) override;  

    /// calls \ref calcVectorField and \ref super::closeTimestep
    void closeTimestep(AdaptInfo *adaptInfo) override;
    
    /// write the solution and the vectorField
    void writeFiles(AdaptInfo *adaptInfo, bool force = false) override;

    // === getting/setting methods ===

    /// return a DOFVector that describes the vectorField P
    DOFVector<WorldVector<double> >* getVectorField()
    { FUNCNAME_DBG("getVectorField()");
      TEST_EXIT_DBG(vectorField != NULL)
	("vectorField is NULL!");
      return vectorField;
    }

    /// implementation of BaseProblem::fillOperators
    void fillOperators() override;
    
    /// used in fillOperators to add the term (grad(P_i), grad(psi))
    virtual void fillLaplacian();

  protected: // variables

    DOFVector<WorldVector<double> >* vectorField;

    void calcVectorField()
    {
      if (self::dow == 1)
	transformDOF(self::prob->getSolution()->getDOFVector(0), 
		    vectorField, new AMDiS::Vec1WorldVec<double>);
      else if (self::dow == 2)
	transformDOF(self::prob->getSolution()->getDOFVector(0), 
		    self::prob->getSolution()->getDOFVector(1), 
		    vectorField, new AMDiS::Vec2WorldVec<double>);
      else if (self::dow == 3)
	transformDOF(self::prob->getSolution()->getDOFVector(0), 
		    self::prob->getSolution()->getDOFVector(1), 
		    self::prob->getSolution()->getDOFVector(2), 
		    vectorField, new AMDiS::Vec3WorldVec<double>);
    }

    double minus1;
    
    double alpha2;
    double alpha4;
    double epsilon;
    double epsInv;
    double K;

    FileVectorWriter *fileWriter;
  };

} // end namespace detail

typedef ::detail::PolarizationField_RB<AMDiS::ProblemStat> PolarizationField_RB;

} }


#include "PolarizationField_RB.hh"

#endif // POLARIZATION_FIELD_RB_H

