/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef CHNS_FUNCTORS_H
#define CHNS_FUNCTORS_H

#include "AMDiS.h"

namespace AMDiS { namespace base_problems {

/** \brief
 * Abstract function for Cahn-Hilliard mobility
 */
class MobilityCH0 : public AbstractFunction<double,double>
{
  public:
    MobilityCH0(double gamma_=1.0, int degree=1) : 
      AbstractFunction<double,double>(degree*4+1), 
      gamma(gamma_),
      delta(1.e-6) { }
    double operator()(const double &ch) const 
    {
      double phase = std::max(0.0, std::min(1.0, ch));
      double mobility = 0.25*sqr(phase)*sqr(phase-1.0);
      return gamma * std::max(mobility, delta);
    }

  protected:
    double gamma;
    double delta;
};

class MobilityCH1 : public AbstractFunction<double,double>
{
  public:
    MobilityCH1(double gamma_=1.0, int degree=1) :
      AbstractFunction<double,double>(degree*4+1),
      gamma(gamma_),
      delta(1.e-6) { }
    double operator()(const double &ch) const
    {
      double phase = std::max(-1.0, std::min(1.0, ch));
      double mobility = 0.25*sqr(sqr(phase)-1.0);
      return gamma * std::max(mobility, delta);
    }

  protected:
    double gamma;
    double delta;
};


class MobilityCH0Diff : public BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >
{
  public:
    MobilityCH0Diff(double factor_=1.0, int degree=1) :
      BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >(degree*3 + (degree-1) + 1),
      factor(factor_) { }
    WorldVector<double> operator()(const double &c, const WorldVector<double>& grdMu) const
    {
      double phase = std::max(0.0, std::min(1.0, c));
      return (factor * (sqr(phase)*phase - 1.5*sqr(phase) + 0.5*phase)) * grdMu;
    }

  protected:
    double factor;
};

class MobilityCH1Diff : public BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >
{
  public:
    MobilityCH1Diff(double factor_=1.0, int degree=1) :
      BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >(degree*3 + (degree-1) + 1),
      factor(factor_) { }
    WorldVector<double> operator()(const double &c, const WorldVector<double>& grdMu) const
    {
      double phase = std::max(-1.0, std::min(1.0, c));
      return (factor * (sqr(phase)*phase - phase)) * grdMu;
    }

  protected:
    double factor;
};

class DoubleWell0 : public AbstractFunction<double,double>
{
  public:
    DoubleWell0(double factor_=1.0, int degree=1) : 
      AbstractFunction<double,double>(3*degree), 
      factor(factor_) { }
      
    double operator()(const double &c) const 
    {
      return factor * (sqr(c)*c - 1.5*sqr(c) + 0.5*c);
    }

  protected:
    double factor;
};

class DoubleWell1 : public AbstractFunction<double,double>
{
  public:
    DoubleWell1(double factor_=1.0, int degree=1) : 
      AbstractFunction<double,double>(3*degree), 
      factor(factor_) { }
      
    double operator()(const double &c) const
    {
      return factor * (sqr(c)*c - c);
    }

  protected:
    double factor;
};

class DoubleWell0Diff : public AbstractFunction<double,double>
{
  public:
    DoubleWell0Diff(double factor_=1.0, int degree=1) : 
      AbstractFunction<double,double>(2*degree), 
      factor(factor_) { }
      
    double operator()(const double &c) const 
    {
      return factor * (3.0*sqr(c) - 3.0*c + 0.5);
    }

  protected:
    double factor;
};

class DoubleWell1Diff : public AbstractFunction<double,double>
{
  public:
    DoubleWell1Diff(double factor_=1.0, int degree=1) : 
      AbstractFunction<double,double>(2*degree), 
      factor(factor_) { }
      
    double operator()(const double &c) const
    {
      return factor * (3.0*sqr(c) - 1.0);
    }

  protected:
    double factor;
};

struct LinearInterpolation : AbstractFunction<double,double>
{
  LinearInterpolation(double v0_, double v1_, double factor_ = 1.0, int degree=1) : 
	AbstractFunction<double,double>(degree), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  double operator()(const double &c) const
  {
	return factor * (v0*(c+1.0) - v1*(c-1.0))/2.0;
  }

private:
  double v0;
  double v1;
  double factor;
};

struct LinearInterpolationFactor : BinaryAbstractFunction<double, double, double>
{
  LinearInterpolationFactor(double v0_, double v1_, double factor_ = 1.0, int degree=1) : 
	BinaryAbstractFunction<double, double, double>(2*degree), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  double operator()(const double &c, const double& u) const
  {
	return factor * u * (v0*(c+1.0) - v1*(c-1.0))/2.0;
  }

private:
  double v0;
  double v1;
  double factor;
};

struct LinearInterpolationDiffFactor : BinaryAbstractFunction<double,double,double>
{
  LinearInterpolationDiffFactor(double v0_, double v1_, double factor_ = 1.0, int degree=1) : 
	BinaryAbstractFunction<double,double,double>(2*degree), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  double operator()(const double &c, const double& z) const
  {
	return factor * z * (v0 - v1)/2.0;
  }

private:
  double v0;
  double v1;
  double factor;
};

struct LinearInterpolationDiffVector : BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >
{
  LinearInterpolationDiffVector(double v0_, double v1_, double factor_ = 1.0, int degree=1) : 
	BinaryAbstractFunction<WorldVector<double>,double,WorldVector<double> >(2*degree-1), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  WorldVector<double> operator()(const double &c, const WorldVector<double>& grdU) const
  {
	return (factor * (v0 - v1)/2.0) * grdU;
  }

private:
  double v0;
  double v1;
  double factor;
};

struct TaylorExpansionFactor : BinaryAbstractFunction<double,double,double>
{
  TaylorExpansionFactor(double v0_, double v1_, double factor_ = 1.0, int degree=1) : 
	BinaryAbstractFunction<double,double,double>(3*degree), 
	v0(v0_), v1(v1_), factor(factor_) { }
      
  double operator()(const double &c, const double& u) const
  {
	return LinearInterpolationFactor(v0, v1, factor)(c, u) + c * LinearInterpolationDiffFactor(v0, v1, factor)(c, u);
  }

private:
  double v0;
  double v1;
  double factor;
};

} }

#endif // CHNS_FUNCTORS_H
