/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef CAHN_HILLIARD_H
#define CAHN_HILLIARD_H

#include "AMDiS.h"
#include "BaseProblem.h"
#include "ProblemStatMassConserve_mm.h"
#include "chns.h"
#include "operations/functors.hpp"
    
namespace AMDiS { namespace base_problems {

namespace detail {
  
  struct c_; // name for the concentration component
  
  class CahnHilliard_mm : public BaseProblem<ProblemStatMassConserve>
  {
  public: // definition of types

    typedef BaseProblem<ProblemStatMassConserve> super;
    typedef CahnHilliard_mm                      self;
    using super::prob;
    
  public: // public methods

    CahnHilliard_mm(const std::string &name_);
    ~CahnHilliard_mm() {};

    void solveInitialProblem(AMDiS::AdaptInfo *adaptInfo) override;

    double getEpsilon() { return eps; }
    int getDoubleWellType() { return doubleWell; }

    void fillOperators() override;
    void fillBoundaryConditions() override {}
    
    void finalizeData() override;

  protected: // protected variables

    bool useMobility;
    bool useReinit;

    unsigned dim;

    int doubleWell;

    double gamma;
    double eps;
    double minusEps;
    double epsInv;
    double minusEpsInv;
    double epsSqr;
    double minusEpsSqr;
  };
  
  
  class MobilityCH0_mm : public FunctorBase //: public AbstractFunction<double,double>
  {
    public:
      MobilityCH0_mm(double gamma_=1.0) : 
	gamma(gamma_),
	delta(1.e-6) { }
	
      typedef double result_type;
      int getDegree(int d0) const { return d0 * 4; }
  
      double operator()(const double &ch) const 
      {
	double phase = std::max(0.0, std::min(1.0, ch));
	double mobility = 0.25*sqr(phase)*sqr(phase-1.0);
	return gamma * std::max(mobility, delta);
      }

    protected:
      double gamma;
      double delta;
  };

  class MobilityCH1_mm : public FunctorBase // : public AbstractFunction<double,double>
  {
    public:
      MobilityCH1_mm(double gamma_=1.0) :
	gamma(gamma_),
	delta(1.e-6) { }
	
      typedef double result_type;
      int getDegree(int d0) const { return d0 * 4; }
      
      double operator()(const double &ch) const
      {
	double phase = std::max(-1.0, std::min(1.0, ch));
	double mobility = 0.25*sqr(sqr(phase)-1.0);
	return gamma * std::max(mobility, delta);
      }

    protected:
      double gamma;
      double delta;
  };

} // end namespace detail

typedef detail::CahnHilliard_mm CahnHilliard_mm;

} }

#include "CahnHilliard_mm.hh"

#endif // CAHN_HILLIARD_H
