/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "AMDiS.h"
#if __cplusplus > 199711L
  #include "CouplingBaseProblem2_cxx11.h"
#else
  #include "CouplingBaseProblem2.h"
#endif
#include "CahnHilliard.h"
// #include "ProblemStatMassConserve2.h"	// using COARSE_RESTRICT
#include "ProblemStatMassConserve.h"		// using multi mesh
#include "NavierStokes_TaylorHood.h"
#include "Refinement.h"
#include "MeshFunction_Level.h"

using namespace AMDiS;

typedef base_problems::detail::CahnHilliard<ProblemStatMassConserve> CHType;
typedef NavierStokes_TaylorHood NSType;

class CahnHilliardNavierStokes : public CouplingBaseProblem< ProblemStat, CHType, NSType >
{
public: // typedefs
  
  typedef CouplingBaseProblem< ProblemStat, CHType, NSType > super;
  
public: // methods

  /// Constructor
  CahnHilliardNavierStokes(std::string name, 
			   CHType &chProb_, 
			   NSType &nsProb_)
  : super(name, chProb_, nsProb_),
    chProb(&chProb_),
    nsProb(&nsProb_),
    refFunction(NULL),
    refinement(NULL),
    minus1(-1.0)
  { }

  /// Destructor
  ~CahnHilliardNavierStokes()
  {
    if (refFunction) { delete refFunction; refFunction = NULL; }
    if (refinement) { delete refinement; refinement = NULL; }
  }

  /// Called before first adaption loop
  virtual void initData()
  {
    refFunction = new PhaseFieldRefinement(chProb->getMesh());
    refinement = new RefinementLevelDOF(chProb->getFeSpace(0), 
					refFunction,
					chProb->getSolution()->getDOFVector(0));
  }

  /// Set initial condition and perform initial refinement
  virtual void solveInitialProblem(AdaptInfo *adaptInfo)
  {    
    // initial refinement
    refinement->refine(0);
    
    // refine until interfaces is resolved
    for (int i = 0; i < 5; ++i) {
      chProb->solveInitialProblem(adaptInfo);
      refinement->refine(5);
    }

    super::solveInitialProblem(adaptInfo);
    
    std::string prefix = "";
    Parameters::get("prefix", prefix);
    
    double c = chProb->getSolution()->getDOFVector(0)->Int();
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    if (MPI::COMM_WORLD.Get_rank() == 0)
#endif
    {
      std::ofstream out(("concentration" + prefix + ".dat").c_str(), ios_base::out);
      out.setf(std::ios::scientific, std::ios::floatfield );
      out.precision(16);
      out << c << "\n";
    }
    // activate mass-conservative coarsening
    refinement->setRefineOperation(adaptInfo, chProb->getProblem());
  }

  /// Called at the end of each timestep.
  virtual void closeTimestep(AdaptInfo *adaptInfo)
  {
    super::closeTimestep(adaptInfo);
    
    std::string prefix = "";
    Parameters::get("prefix", prefix);
    
    double c = chProb->getSolution()->getDOFVector(0)->Int();
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    if (MPI::COMM_WORLD.Get_rank() == 0)
#endif
    {
      std::ofstream out(("concentration" + prefix + ".dat").c_str(), ios_base::app);
      out.setf(std::ios::scientific, std::ios::floatfield );
      out.precision(16);
      out << c << "\n";
    }
    
    refinement->refine(1);
    // checkMeshChange
  }
  
  virtual void fillCouplingOperators()
  {
    double M0 = 1.0;
    Parameters::get(name + "->M0",M0);
    
    DOFVector<double>* c = chProb->getSolution()->getDOFVector(0);
    DOFVector<double>* mu = chProb->getSolution()->getDOFVector(1);
    
    // < grad(mu) * c , theta >
    for (size_t i = 0; i < dow; i++) {
      Operator *opCGradMu = new Operator(nsProb->getFeSpace(i), 
					 chProb->getFeSpace(0));
      addZOT(opCGradMu, M0 * valueOf(mu) * derivativeOf(c, i) );
      nsProb->getProblem(0)->addVectorOperator(opCGradMu, i);
    }
    
    // < - u * c , grad(theta) >
    for (size_t i = 0; i < dow; i++) {
      Operator *opUGradC = new Operator(chProb->getFeSpace(1), 
					chProb->getFeSpace(0));      
      addFOT(opUGradC, -valueOf(nsProb->getSolution()->getDOFVector(i)), i, GRD_PSI );
      chProb->getProblem()->addMatrixOperator(opUGradC, 1, 0);
    }
  }

protected: // variables

  CHType *chProb;
  NSType *nsProb;
  PhaseFieldRefinement* refFunction;
  RefinementLevelDOF *refinement;

  double minus1;
  double concentration;
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);  
  Timer t;

  CHType chProb("ch");  
  NSType nsProb("ns");

  CahnHilliardNavierStokes chns("chns", chProb, nsProb);
  chns.initialize(INIT_ALL);
  chns.initTimeInterface();

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", chns.getNumComponents());
  AdaptInstationary adaptInstat("adapt", chns, adaptInfo, chns, adaptInfo);

  int error_code = adaptInstat.adapt(); 

  MSG("elapsed time= %d sec\n", t.elapsed());
  AMDiS::finalize();

  return error_code;
}

