#include "AMDiS.h"
#include "MyProblemInstat.h"
#include "Anisotropy.h"

using namespace AMDiS;
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
using namespace AMDiS::Parallel;
#endif
using namespace AMDiS::io;

class DoublewellExpl : public AbstractFunction<double, double>
{
public:
  DoublewellExpl()
    : AbstractFunction<double, double>(3)
  {}

  double operator()(const double& u) const
  {    
    return 2.0 * u * u * u;
  }
};

class DoublewellImpl : public AbstractFunction<double, double>
{
public:
  DoublewellImpl()
    : AbstractFunction<double, double>(2)
  {}

  double operator()(const double& u) const
  {
    return 3.0 * u * u - 1.0;
  }
};

class Karma : public AbstractFunction<double, double>
{
public:
  Karma()
    : AbstractFunction<double, double>(4)
  {}

  double operator()(const double& u) const
  {
    return pow((1.0 - u * u), 2.0);
  }
};

using namespace std;
typedef vector<vector<double> > DataVec;

void scalMesh(Mesh *mesh, double scalFactorX, double scalFactorY, double scalFactorZ)
{
  for (std::deque<MacroElement*>::iterator it = mesh->firstMacroElement();
       it !=  mesh->endOfMacroElements(); ++it) {
    for (int i = 0; i < mesh->getGeo(VERTEX); i++) {
      WorldVector<double> &x = (*it)->getCoord(i);
      
      if (scalFactorX != 0.0) 
	x[0] *= scalFactorX;
      
      if (scalFactorY != 0.0) 
	x[1] *= scalFactorY;
      
      if (scalFactorZ != 0.0) 
	x[2] *= scalFactorZ;
    }
  }
}


int main(int argc, char** argv)
{
  FUNCNAME("main");

  Timer t;

  AMDiS::init(argc, argv);

  ProblemStat dendSpace("dendrites->space");
  dendSpace.initialize(INIT_ALL);

  MyProblemInstat probInstat("dendrites", dendSpace);
  probInstat.initialize(INIT_ALL);

  // create adapt info
  AdaptInfo adaptInfo("dendrites->adapt", dendSpace.getNumComponents());

  // create initial adapt info
  AdaptInfo adaptInfoInitial("dendrites->initial->adapt", dendSpace.getNumComponents());

  // create instationary adapt
  AdaptInstationary adaptInstat("dendrites->adapt",
				dendSpace,
				adaptInfo,
				probInstat,
				adaptInfoInitial);

  // === Scal mesh ===

  double scalFactorX = 0.0;
  Parameters::get("dendritesMesh->scal x", scalFactorX);

  double scalFactorY = 0.0;
  Parameters::get("dendritesMesh->scal y", scalFactorY);

  double scalFactorZ = 0.0;
  Parameters::get("dendritesMesh->scal z", scalFactorZ);
  
  if (!dendSpace.isDeserialized()) {
    if (scalFactorX != 0.0 || scalFactorY != 0.0 || scalFactorZ != 0.0) {
      scalMesh(dendSpace.getMesh(0), scalFactorX, scalFactorY, scalFactorZ);
      
      if (dendSpace.getMesh(0) != dendSpace.getMesh(1))
	scalMesh(dendSpace.getMesh(1), scalFactorX, scalFactorY, scalFactorZ);
      
      if (scalFactorZ == 0.0)
	MSG("Mesh scaled with %f x %f\n", scalFactorX, scalFactorY);
      else
	MSG("Mesh scaled with %f x %f x %f\n", scalFactorX, scalFactorY, scalFactorZ);
    } 
  }


 
  double thermalDif = 0.0;
  Parameters::get("dendrites->karma model->D", thermalDif);

  double a2 = 0.6267;
  double lambda = thermalDif / a2;

  double gamma = 0.0;
  Parameters::get("dendrites->karma model->gamma", gamma);

  double rhoInitial = 0.0;
  Parameters::get("dendrites->karma model->rho initial", rhoInitial);

  MSG("Karma model parameters:\n");
  MSG("    D     = %f\n", thermalDif);
  MSG("    gamma = %f\n", gamma);
  MSG("    Delta = %f\n", rhoInitial);
  MSG("\n");

  DOFVector<double> *oldPhi = probInstat.getOldSolution()->getDOFVector(0);

  AbstractFunction<double, WorldVector<double> > *anisoSquare;
  AbstractFunction<WorldMatrix<double>, WorldVector<double> > *anisoLaplace;

  switch (dendSpace.getMesh(0)->getDim()) {
  case 2:
    anisoSquare = new Aniso2d_Square(gamma);
    anisoLaplace =  new Aniso2d_Laplace(gamma);
    break;
  case 3:
    anisoSquare = new Aniso3d_Square(gamma);
    anisoLaplace = new Aniso3d_Laplace(gamma);
    break;
  default:
    ERROR_EXIT("This should not happen!\n");
  }

  NullFct nullFct;

  ///

  Operator opPhiTau(dendSpace.getFeSpace(0), dendSpace.getFeSpace(0));
  opPhiTau.addZeroOrderTerm(new FctGradient_ZOT(oldPhi, anisoSquare));
  opPhiTau.setUhOld(probInstat.getOldSolution()->getDOFVector(0));
  dendSpace.addMatrixOperator(opPhiTau, 0, 0, probInstat.getInvTau(), probInstat.getInvTau());
  dendSpace.addVectorOperator(opPhiTau, 0, probInstat.getInvTau(), probInstat.getInvTau());

  Operator opDoublewellImpl(dendSpace.getFeSpace(0), dendSpace.getFeSpace(0));
  opDoublewellImpl.addZeroOrderTerm(new VecAtQP_ZOT(oldPhi, new DoublewellImpl));
  dendSpace.addMatrixOperator(opDoublewellImpl, 0, 0);

  Operator opDoublewellExpl(dendSpace.getFeSpace(0));
  opDoublewellExpl.addZeroOrderTerm(new VecAtQP_ZOT(oldPhi, new DoublewellExpl));
  dendSpace.addVectorOperator(opDoublewellExpl, 0);

  Operator opKarmaFct(dendSpace.getFeSpace(0), dendSpace.getFeSpace(1));
  opKarmaFct.addZeroOrderTerm(new VecAtQP_ZOT(oldPhi, new Karma));
  dendSpace.addMatrixOperator(opKarmaFct, 0, 1, &lambda, &lambda);

  Operator opAnisoLaplace(dendSpace.getFeSpace(0), dendSpace.getFeSpace(0));
  opAnisoLaplace.addSecondOrderTerm(new MatrixGradient_SOT(oldPhi, anisoLaplace, &nullFct));
  dendSpace.addMatrixOperator(opAnisoLaplace, 0, 0);


  ///

  Operator opLaplace(dendSpace.getFeSpace(1), dendSpace.getFeSpace(1));
  opLaplace.addSecondOrderTerm(new Simple_SOT(&thermalDif));
  dendSpace.addMatrixOperator(opLaplace, 1, 1);

  Operator opUTau(dendSpace.getFeSpace(1), dendSpace.getFeSpace(1));
  opUTau.addZeroOrderTerm(new Simple_ZOT);
  opUTau.setUhOld(probInstat.getOldSolution()->getDOFVector(1));
  dendSpace.addMatrixOperator(opUTau, 1, 1, probInstat.getInvTau(), probInstat.getInvTau());
  dendSpace.addVectorOperator(opUTau, 1, probInstat.getInvTau(), probInstat.getInvTau());

  
  Operator opPhiUTau(dendSpace.getFeSpace(1), dendSpace.getFeSpace(0));
  opPhiUTau.addZeroOrderTerm(new Simple_ZOT(-0.5));
  opPhiUTau.setUhOld(probInstat.getOldSolution()->getDOFVector(0));
  dendSpace.addMatrixOperator(opPhiUTau, 1, 0, probInstat.getInvTau(), probInstat.getInvTau());
  dendSpace.addVectorOperator(opPhiUTau, 1, probInstat.getInvTau(), probInstat.getInvTau());

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  MeshDistributor::globalMeshDistributor->addInterchangeVector(probInstat.getOldSolution());
  MeshDistributor::globalMeshDistributor->addInterchangeVector(dendSpace.getSolution());
#endif

  // ===== start adaption loop =====
  int errorCode = adaptInstat.adapt();

  
  printf("whole process needs %.5f seconds\n", t.elapsed()); 

  
  bool isWrite = false; string phiname, rhoname;
  Parameters::get("dendrites->write arh file", isWrite);
  Parameters::get("dendrites->write phi arh", phiname);
  Parameters::get("dendrites->write rho arh", rhoname);
  
  if (isWrite) {
    io::Arh2Writer::writeFile(dendSpace.getSolution(0), phiname.c_str());
    io::Arh2Writer::writeFile(dendSpace.getSolution(1), rhoname.c_str());
  }
  
  AMDiS::finalize();

  return errorCode;   
}
