/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

namespace AMDiS { namespace extensions { 
  
void MTLPreconPfcDiag::init(const BlockMatrix& A_, const MTLMatrix& fullMatrix_)
{
  assert(tau != NULL);
  
  super::init(A_, fullMatrix_);
  
  typedef typename mtl::Collection<MTLMatrix>::size_type size_type;
  typedef typename mtl::Collection<MTLMatrix>::value_type value_type;
  size_type n = num_rows(getM());
  
  double delta = std::sqrt(getTau());
  
  // helper-matrix MpL = M + delta*L
  MpL.change_dim(n, n);
  MpL = getM() + delta*getL();

  // helper-matrix MpL = M + sqrt(delta)*L
  S.change_dim(n, n);
  MinvL.change_dim(n, n);
  MinvL = diagonal(diagonal(getM()));
  invert_diagonal(MinvL);
  MinvL *= getL();
  
  S = getM() - ( 2.0 * delta ) * getL() + delta * getL() * MinvL;
  
  bool useDirectSolverM = false, 
       useDirectSolverMpL = false, 
       useDirectSolverMpL2 = false;
  Parameters::get("precon_pfc_M->use direct solver", useDirectSolverM);
  Parameters::get("precon_pfc_MpL->use direct solver", useDirectSolverMpL);
  Parameters::get("precon_pfc_MpL2->use direct solver", useDirectSolverMpL2);
  
  PId = new itl::pc::identity<MTLTypes::MTLMatrix, double>(getM());
  
  if (useDirectSolverM) {
    try {
      solverM = new mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix>(getM(), UMFPACK_STRATEGY_SYMMETRIC, 0.7);
    } catch (mtl::matrix::umfpack::error& e) {
      ERROR_EXIT("UMFPACK_ERROR(M, %d) = %s\n", e.code, e.what());
    }
  } else
    PDiagM = new itl::pc::diagonal<MTLTypes::MTLMatrix, double>(getM());
  if (useDirectSolverMpL) {
    try {
      solverMpL = new mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix>(MpL, UMFPACK_STRATEGY_SYMMETRIC, 0.7);
    } catch (mtl::matrix::umfpack::error& e) {
      ERROR_EXIT("UMFPACK_ERROR(MpL, %d) = %s\n", e.code, e.what());
    }
  } else
    PDiagMpL = new itl::pc::diagonal<MTLTypes::MTLMatrix, double>(MpL);
  if (useDirectSolverMpL2) {
    try {
      solverS = new mtl::matrix::umfpack::solver<MTLTypes::MTLMatrix>(S, UMFPACK_STRATEGY_SYMMETRIC, 0.7);
    } catch (mtl::matrix::umfpack::error& e) {
      ERROR_EXIT("UMFPACK_ERROR(MpL2, %d) = %s\n", e.code, e.what());
    }
  } else
    PDiagS = new itl::pc::diagonal<MTLTypes::MTLMatrix, double>(S);
  
  // temporary variables
  y0.change_dim(num_rows(getM()));
  y1.change_dim(num_rows(getM()));
  tmp.change_dim(num_rows(getM()));
}


void MTLPreconPfcDiag::solve(const MTLVector& b, MTLVector& x) const
{ FUNCNAME("MTLPreconPfcDiag::solve()");

  x.change_dim(num_rows(b));
  
  const MTLVector b0(b[rows[0]]);
  const MTLVector b1(b[rows[1]]);
  const MTLVector b2(b[rows[2]]);
  
  MTLVector x0(x[rows[0]]);
  MTLVector x1(x[rows[1]]);
  MTLVector x2(x[rows[2]]);
  
  double delta = std::sqrt(getTau());
  
  solveM(y0, b0);		// M*y0 = b0
  y1 = getL() * y0;		// y1 = K*y0
  tmp = b1 - getTau()*y1;	// tmp := b1 - tau*y1

  solveMpL(y1, tmp);		// (M + delta*K) * y1 = tmp
  x0 = y0 + (1.0/delta)*y1;	// x0 = y0 + (1/delta)*y1

  tmp = getM() * y1;		// tmp := M*y1
  solveS(x1, tmp);		// (M+eps*sqrt(delta)K) * x1 = tmp

  x0-= (1.0/delta)*x1;		// x0 = x0 - (1/delta)*x1 = y0 + (1/delta)*(y1 - x1)

  tmp = b2 - getL() * x1;	// tmp := b2 - K*x1
  solveM(x2, tmp);
}

} }
