/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#include "PetscPreconPfc.h"

namespace AMDiS {

  using namespace std;
  
  /// solve Pfc Preconditioner
  PetscErrorCode pcPfcShell(PC pc, Vec b, Vec x) // solve Px=b
  {       
    void *ctx;
    PCShellGetContext(pc, &ctx);
    PfcData* data = static_cast<PfcData*>(ctx);
    
    Vec b1, b2, b3, x1, x2, x3;
    VecNestGetSubVec(b, 0, &b1);
    VecNestGetSubVec(b, 1, &b2);
    VecNestGetSubVec(b, 2, &b3);

    VecNestGetSubVec(x, 0, &x1);
    VecNestGetSubVec(x, 1, &x2);
    VecNestGetSubVec(x, 2, &x3);


    // Hilfsvariablen
    Vec y1, y2, tmp;
    VecDuplicate(b1, &y1);
    VecDuplicate(b1, &y2);
    VecDuplicate(b1, &tmp);

    KSPSolve(data->kspM, b1, y1); 			// M*y1 = b1    
    MatMult(data->matK, y1, tmp); 			// tmp := K*y1
    VecAYPX(tmp, -sqr(data->delta), b2);			// tmp := b2 - tau*tmp
    
    KSPSolve(data->kspMpK, tmp, y2); 			// (M+sqrt(tau)K)*y2 = tmp
    
    MatMult(data->matM, y2, tmp);			// tmp := M*y2
    KSPSolve(data->kspMpK2, tmp, x2); 			// MpK2*x2 = tmp
    MatMult(data->matM, x2, tmp);			// tmp := M*x2
    KSPSolve(data->kspMpK2, tmp, x2); 			// MpK2*x2 = tmp

    VecCopy(x2, x1); 					// x1 := x2
    VecAXPBYPCZ(x1, 1.0, 1.0/(data->delta), -1.0/(data->delta), y1, y2);	// x1 = 1*y1 + factor*y2 - factor*x1

    MatMult(data->matK, x2, tmp); 			// tmp := K*x2
    VecAYPX(tmp, -1.0, b3);				// tmp := b3 - tmp
    
    KSPSolve(data->kspM, tmp, x3); 			// M*x3 = tmp    

    VecDestroy(&y1);
    VecDestroy(&y2);
    VecDestroy(&tmp);
    
    return 0;
  }
  

  void PetscPreconPfc::init(PC pc, const SolverMatrix<Matrix<DOFMatrix*> >&, const MatrixType& A)
  { FUNCNAME("PetscPreconPfc::init()");
    
    assert(tau != NULL);
    
    // init shell preconditioner
    PCSetType(pc, PCSHELL);
    PCShellSetApply(pc, pcPfcShell);
    PCShellSetContext(pc, &data);
    
  
    double delta = std::sqrt((*tau) * M0);
    
    MatNestGetSubMat(A.matrix, 2, 2, &data.matM);
    MatNestGetSubMat(A.matrix, 2, 1, &data.matK);
    
    MatDuplicate(data.matM, MAT_COPY_VALUES, &MpK);
    MatAXPY(MpK, delta, data.matK, SAME_NONZERO_PATTERN);
    
    MatDuplicate(data.matM, MAT_COPY_VALUES, &MpK2);
    MatAXPY(MpK2, std::sqrt(delta), data.matK, SAME_NONZERO_PATTERN);
    
    // init sub-solvers
    Runner::createSubSolver(data.kspM, data.matM, "mass_");
    Runner::createSubSolver(data.kspMpK, MpK, "MpK_");
    Runner::createSubSolver(data.kspMpK2, MpK2, "MpK2_");
    
    data.delta = delta;
    data.tau = *tau;
    

    int nIterM=5, nIterMpL=20, nIterMpL2=10;
    double tolM=PETSC_DEFAULT, tolMpL=PETSC_DEFAULT, tolMpL2=PETSC_DEFAULT;
    double rtolM=PETSC_DEFAULT, rtolMpL=PETSC_DEFAULT, rtolMpL2=PETSC_DEFAULT;
    Parameters::get("precon_pfc_M->max iteration", nIterM);
    Parameters::get("precon_pfc_MpL->max iteration", nIterMpL);
    Parameters::get("precon_pfc_MpL2->max iteration", nIterMpL2);
    Parameters::get("precon_pfc_M->tolerance", tolM);
    Parameters::get("precon_pfc_MpL->tolerance", tolMpL);
    Parameters::get("precon_pfc_MpL2->tolerance", tolMpL2);
    Parameters::get("precon_pfc_M->relative tolerance", rtolM);
    Parameters::get("precon_pfc_MpL->relative tolerance", rtolMpL);
    Parameters::get("precon_pfc_MpL2->relative tolerance", rtolMpL2);
    
    bool useAMG = false;
    Parameters::get("precon_pfc_MpL2->use AMG", useAMG);

    Runner::setSolver(data.kspM, "M_", KSPCG, PCJACOBI, rtolM, tolM, nIterM);
    Runner::setSolver(data.kspMpK, "MpK_", KSPCG, PCJACOBI, rtolMpL, tolMpL, nIterMpL);
    if (!useAMG) {
      Runner::setSolver(data.kspMpK2, "MpK2_", KSPCG, PCJACOBI, rtolMpL2, tolMpL2, nIterMpL2);
    } else {
      Runner::setSolver(data.kspMpK2, "MpK2_", KSPRICHARDSON, PCHYPRE, rtolMpL2, tolMpL2, nIterMpL2);
    }
  }


  void PetscPreconPfc::exit()
  {    
    MatDestroy(&MpK);
    MatDestroy(&MpK2);
    
    KSPDestroy(&data.kspM);
    KSPDestroy(&data.kspMpK);
    KSPDestroy(&data.kspMpK2);
  }
}
